/* -*- mode: c++; c-basic-offset: 4; -*- */
//
//    Multiple streams of data access implementation
//
#include "MultiDacc.hh"
#include <cstdio>
#include <iostream>
#include <algorithm>

using namespace std;

//  dTcontig is the maximum acceptable discontinuity in nsec.
static const int dTcontig = 1;

//======================================  Default (Null) constructor.
MultiDacc::MultiDacc(void)
    : mStride(0.0), mFillStride(0.0), mTimeout(-1.0)
{
}

//======================================  Construct and connect to source.
MultiDacc::MultiDacc(const std::string& Source, const Time& STime)
    : mStride(0.0), mFillStride(0.0), mTimeout(-1.0)
{
    addMulti(Source);
}

//======================================  Destructor.
MultiDacc::~MultiDacc(void) {
    close();
    mChanList.clear();
}

//======================================  Add multiple frame sources.
void
MultiDacc::addMulti(const string& name) {
    size_t n0 = mList.size();
    mList.addMulti(name.c_str());
    for(size_t i=n0; i<mList.size(); i++) {
	Dacc* newDacc = new Dacc();
	newDacc->addPathList(mList.getList(i).c_str());
	mIn.push_back(dacc_status(newDacc));
    }
}

//======================================  Add a single frame source.
void
MultiDacc::addSingle(const string& name) {
    mList.addSingle(name.c_str());
    Dacc* newDacc = new Dacc();
    if (name[0] == '=') newDacc->addPath(name.substr(1));
    else                newDacc->addPathList(name);
    mIn.push_back(dacc_status(newDacc));
}
//======================================  Close all frame readers
void 
MultiDacc::channel_id(const string& chanid, int& daccid, string& chan) const {
    size_t inx = chanid.find_first_of('/');
    if (inx == string::npos) {
	chan = chanid;
	daccid = getDaccIndex(chan);
    } else if (chanid[0] >= '0' && chanid[0] <= '9') {
	daccid = strtoul(chanid.c_str(), 0, 0);
	chan = chanid.substr(inx+1);
    } else {
	daccid = frame_type(chanid.substr(0, inx));
	chan = chanid.substr(inx+1);
    }
}

//======================================  Close all frame readers
void
MultiDacc::close(void) {
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	(*i)->close();
    }
}

//======================================  Open a frame file on all streams
int
MultiDacc::open() {
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	int j=(*i)->open();
	if(j != 0) return j;
    }
    return 0;
}

//======================================  Set the allow_missing_frame flag
void
MultiDacc::setAllowMissingFrame(int id, bool yorn) {
    mIn[id].allow_missing_frame(yorn);
}
	
//======================================  Set the buffer count on all streams
void
MultiDacc::setBuffer(int nBuf) {
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	(*i)->setBuffer(nBuf);
    }
}

//======================================  Set the debug level on all streams
void
MultiDacc::setDebug(int nDbg) {
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	(*i)->setDebug(nDbg);
    }
}

//======================================  Set flag if one or more streams online
void
MultiDacc::setIfOnline(void) {
    mIsOnline = any_of(mIn.begin(), mIn.end(),
		       [](const dacc_status& x) {return x->isOnline();}); 
}

//======================================  Set the ignore flag on all streams
void
MultiDacc::setIgnoreMissingChannel(bool yn) {
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	(*i)->setIgnoreMissingChannel(yn);
    }
}

//======================================  Set the nowait flag on all streams
void
MultiDacc::setNoWait(bool now) {
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	(*i)->setTimeout(now ? 0.0 : mTimeout);
    }
}

//======================================  Set the stride on all streams
void
MultiDacc::setStride(Interval Dt) {
    mStride = Dt;
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	(*i)->setStride(mStride);
    }
}

//======================================  Set the stride on all streams
void
MultiDacc::setTimeout(double Dt) {
    mTimeout = Dt;
}

//======================================  Get debug level
int
MultiDacc::getDebug(void) {
    if (mIn.empty()) return 0;
    return mIn[0]->getDebug();
}

//======================================  Get number of frames read
long
MultiDacc::getTotalFrames(void) const {
    if (mIn.empty()) return 0;
    return mIn[0]->getTotalFrames();
}

//======================================  Get current time (or Time(0))
Time
MultiDacc::getCurrentTime(void) const {
    if (mIn.empty()) return Time(0);
    return mIn[0]->getCurrentTime();
}

//======================================  Get fill stride
Interval
MultiDacc::getFillStride(void) const {
    return mFillStride;
}

//======================================  Get fill time (or Time(0))
Time
MultiDacc::getFillTime(void) const {
    if (mIn.empty()) return Time(0);
    return mIn[0]->getFillTime();
}

//======================================  Get stride (or 0)
Interval
MultiDacc::getStride(void) const {
    return mStride;
}

//======================================  Test the allow_missing_frame flag
bool
MultiDacc::ignoreMissingFrame(int id) const {
    return mIn[id].ignore_missing_frame();
}

//======================================  Add a channel to the request list.
void 
MultiDacc::addChannel(const string& Name, int id, int Decimate,
		      TSeries **TSptr) {
    int daccid;
    string channame;
    channel_id(Name, daccid, channame);

    chan_iter i=findChannel(channame);
    if (i != mChanList.end()) {
	if (getDebug()) cout << "Existing channel " << Name 
			     << " entry replaced." << endl;
	mChanList.erase(i);
    }
    
    if (id < 0) id = daccid;
    mChanList.push_back(ChannelIndex(channame, id));
    if (id < 0 || id >= int(mIn.size())) {
	cerr << "MultiDacc: failed to add channel " << Name << endl;
    } else {
	mIn[id]->addChannel(channame, Decimate, TSptr);
    }
}

//======================================  Add a channel to the request list.
void 
MultiDacc::addFSeries(const string& Name, int id, FSeries **FSptr) {
    int daccid;
    string channame;
    channel_id(Name, daccid, channame);

    chan_iter i=findChannel(channame);
    if (i != mChanList.end()) {
	if (getDebug()) cout << "Existing channel " << Name 
			     << " entry replaced." << endl;
	mChanList.erase(i);
    }
    
    if (id < 0) id = daccid;
    mChanList.push_back(ChannelIndex(channame, id));
    if (id < 0 || id >= int(mIn.size())) {
	cerr << "MultiDacc: failed to add channel " << Name << endl;
    } else {
	mIn[id]->addFSeries(channame, FSptr);
    }
}

//======================================  Add a channel to the request list.
void 
MultiDacc::addRaw(const string& Name, int id, int Decimate, TSeries **TSptr) {
    int daccid;
    string channame;
    channel_id(Name, daccid, channame);

    chan_iter i=findChannel(channame);
    if (i != mChanList.end()) {
	if (getDebug()) cout << "Existing channel " << Name 
			     << " entry replaced." << endl;
	mChanList.erase(i);
    }
    
    if (id < 0) id = daccid;
    mChanList.push_back(ChannelIndex(channame, id));
    if (id < 0 || id >= int(mIn.size())) {
	cerr << "MultiDacc: failed to add channel " << Name << endl;
    } else {
	mIn[id]->addRaw(channame, Decimate, TSptr);
    }
}

//======================================  Add a channel to the request list.
void 
MultiDacc::addProcessed(const string& Name, int id, int Decimate,
			TSeries **TSptr) {
    int daccid;
    string channame;
    channel_id(Name, daccid, channame);

    chan_iter i=findChannel(channame);
    if (i != mChanList.end()) {
	if (getDebug()) cout << "Existing channel " << Name 
			     << " entry replaced." << endl;
	mChanList.erase(i);
    }
    
    if (id < 0) id = daccid;
    mChanList.push_back(ChannelIndex(channame, id));
    if (id < 0 || id >= int(mIn.size())) {
	cerr << "MultiDacc: failed to add channel " << Name << endl;
    } else  {
	mIn[id]->addProcessed(channame, Decimate, TSptr);
    }
}

//======================================  Add a channel to the request list.
void 
MultiDacc::addSimulated(const string& Name, int id, int Decimate,
			TSeries **TSptr) {
    int daccid;
    string channame;
    channel_id(Name, daccid, channame);

    chan_iter i=findChannel(channame);
    if (i != mChanList.end()) {
	if (getDebug()) cout << "Existing channel " << Name 
			     << " entry replaced." << endl;
	mChanList.erase(i);
    }
    
    if (id < 0) id = daccid;
    mChanList.push_back(ChannelIndex(channame, id));
    if (id < 0 || id >= int(mIn.size())) {
	cerr << "MultiDacc: failed to add channel " << Name << endl;
    } else  {
	mIn[id]->addSimulated(channame, Decimate, TSptr);
    }
}

//======================================  See if channel was requested.
bool 
MultiDacc::isChannelRead(const string& Name) const {
    return findChannel(Name) != mChanList.end();
}

//======================================  List requested channels
ostream&
MultiDacc::list(ostream& out) const {
    for (size_t id=0; id<mIn.size(); id++) {
	out << "Data stream " << id << ": frame_type: " << frame_name(id)
	    << endl;
	mIn[id]->list(out);
    }
    return out;
}

//======================================  Frame type of the specified stream
std::string
MultiDacc::frame_name(int id) const {
    if (id < 0 || id >= int(mIn.size())) return "";
    string framepath = mIn[id]->refList().first();
    string::size_type inx = framepath.rfind("/");
    if (inx != string::npos) framepath.erase(0, inx+1);
    inx = framepath.find("-");
    if (inx != string::npos) {
	string::size_type inx2 = framepath.find("-", inx+1);
	if (inx2 != string::npos) inx = inx2;
    }
    if (inx != string::npos) framepath.erase(inx);
    return framepath;
}

//======================================  Stream with the specified frame type
int
MultiDacc::frame_type(const std::string& frameType) const {

   //--------------------------------  Find an input stream ID.
   size_t N = mIn.size();
   for (size_t i=0; i < N; ++i) {
       if (frame_name(i) == frameType) return i;
   }
   return -1;
}

//======================================  Find channel in request list.
MultiDacc::const_chan_iter 
MultiDacc::findChannel(const std::string& name) const {
    size_t inx = name.find_first_of('/');
    string raw = (inx == string::npos) ? name : name.substr(inx+1);
    for (const_chan_iter i=mChanList.begin() ; i != mChanList.end() ; i++) {
	if (i->EqName(raw)) return i;
    }
    return mChanList.end();
}

MultiDacc::chan_iter 
MultiDacc::findChannel(const string& name) {
    size_t inx = name.find_first_of('/');
    string raw = (inx == string::npos) ? name : name.substr(inx+1);
    for (chan_iter i=mChanList.begin() ; i != mChanList.end() ; i++) {
	if (i->EqName(raw)) return i;
    }
    return mChanList.end();
}

//======================================  Remove channel from request list.
void 
MultiDacc::rmChannel(const string& Name) {
    int index=getDaccIndex(Name);
    if (index != -1) {
	mIn[index]->rmChannel(Name);
    }
    chan_iter i=findChannel(Name);
    if (i != mChanList.end()) {
	mChanList.erase(i);
    }
}

//======================================  Synchronize the data streams
int 
MultiDacc::synch(void) {
    int rc= 0;
    
     //----------------------------------  See how they stack up
    Time tEarly(0), tLate(0), now(Now());
    if (getDebug()) cerr << "Start Synch:";
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	int istr = i - mIn.begin();
	if (i->ignore_missing_frame()) {
	    if ((*i)->isOnline() && mTimeout > 0) {
		rc = i->tseek(tLate, mTimeout);
		if (rc && getDebug()) cerr << " t[" << i - mIn.begin()
					   << "] ignored (rc " << rc << ")";
		if (rc == -10) {
		    rc = 0;
		    continue;
		}
	    } else {
		rc = (*i)->synch();
	    }
	    if (rc) break;
	    Time ti = (*i)->getCurrentTime();
	    if (getDebug()) cerr << " t[" << istr << "]=" << ti.getS();
	    if (!tEarly || ti < tEarly) tEarly = ti;
	} else {
	    rc = (*i)->synch();
	    if (rc) break;
	    Time ti = (*i)->getCurrentTime();
	    if (getDebug()) cerr << " t[" << istr << "]=" << ti.getS();
	    if (!tEarly || ti < tEarly) tEarly = ti;
	    if (ti > tLate) tLate = ti;
	}
    }
    if (getDebug()) cerr << endl;
    if (rc && rc != -10) return rc;

    //----------------------------------  Seek the latest required.
    rc = 0;
    mSynchTime = tLate;
    while (! Almost(tEarly, tLate) ) {
	tEarly = Time(0);
	if (getDebug()) cerr << "Synch to: " << mSynchTime.getS()
			     << endl;
	for(auto i=mIn.begin(); i != mIn.end(); ++i) {
	    //--------------------------  release if frame isn't required
	    if (i->ignore_missing_frame()) {
		if ((*i)->isOnline() && mTimeout > 0) {
		    (*i)->setTimeout(mTimeout);
		}
		rc = (*i)->seek(mSynchTime);
		(*i)->setTimeout(-1);
		
		if (rc && getDebug()) cerr << " t[" << i - mIn.begin()
					   << "] ignored (rc " << rc << ")";
		
		if (rc == -10) {
		    rc = 0;
		    continue;
		}
		if (rc) break;
		Time ti = (*i)->getCurrentTime();
		if (getDebug()) cerr << " t[" << i - mIn.begin() << "]="
				     << ti.getS();
	    }
	    
	    //--------------------------- get start times for required frames
	    else {
		int rc = (*i)->seek(mSynchTime);
		if (rc) {
		    if (getDebug()) cerr << " t[" << i - mIn.begin()
					 << "] ignored (rc " << rc << ")";
		    break;
		}
		Time ti = (*i)->getCurrentTime();
		if (getDebug()) cerr << " t[" << i - mIn.begin() << "]="
				     << ti.getS();
		if (ti > tLate) tLate = ti;
		if (!tEarly || ti < tEarly) tEarly =ti;
	    }
	}
	if (getDebug()) cerr << endl;
	if (rc) return rc;
	mSynchTime = tLate;
    }
    return 0;
}

//======================================  Fill requested series containers.
//    Error codes:
//      0  Successful completion
//     -1  Frame start not contiguous to previous data
//     -2  Sample rate isn't compatible with previous data.
//     -3  Requested data not found in current frame 
//     -4  Error reading frame
//     -5  Frame data are not self-consistent.
//     -6  TSeries is not allocated.
//     -7  Unsupported data type
//     -8  Interrupted by a signal or online data not ready.
//     -9  Invalid data (NaN, Inf or Unnormalized).
//     -10 Data timed out.
int 
MultiDacc::fillData(Interval Stride, bool start) {
    setIfOnline();
    //----------------------------------  Synch and initialize the status at
    //                                    the start of a fill.
    int rc = 0;
    if (start) {
	for_each(mIn.begin(), mIn.end(),
		 [](dacc_status& x){x.set_state(1);}
		 );
	
	if (!Stride) 
	    if (!mStride) mFillStride = 1.0;
	    else          mFillStride = mStride;
	else              mFillStride = Stride;
 
	zeroChans(mFillStride);
	rc = synch();
	if (rc) return rc;
    }

    //----------------------------------  Loop over first and subsequent
    //                                    interrupted data streams.
    bool first = true;
    for (auto i=mIn.begin(); i != mIn.end(); i++) {
	int rci = i->state();
	if (rci > 0 || rci == -8) {
	    //--------------------------  Wait for data in first stream only
	    if (mIsOnline) {
		if (first) {
		    rci = i->tseek(mSynchTime, mTimeout);
		}
		//----------------------  See if data are available. Bail if not.
		else {
		    rci = i->tseek(mSynchTime, 0.0);
		    if (rci == -10) rci = -8;
		}
		first = false;
	    }
	    
	    //--------------------------  If not online, synch, test for correct
	    else {
		rci = (*i)->synch();
	    }
	    
	    if (i->ignore_missing_frame()) {
		if (rci == -10) continue;
		if (!rci && (*i)->getCurrentTime() > mSynchTime) continue;
	    }
	    if (rci) return rci;
	    
	    rci = (*i)->fillData(mFillStride, rci > 0);
	    i->set_state(rci);
	    
	    //---------------------------  Set the return code
	    if (rci) return rci;
	}
    }

    //----------------------------------  Set start time for missing streams
    for (auto i=mIn.begin(); i != mIn.end(); i++) {
	if (i->state() == 1) {
	    if (getDebug()) 
		cerr << "fillData: set Null stream " << i-mIn.begin() << " t0 = "
		     << mSynchTime << " dt = " << mFillStride << endl;
	    (*i)->nullChans(mSynchTime, Stride);
	}
    }
    return 0;
}

//======================================  Clear all channel output series.
void
MultiDacc::zeroChans(Interval dT) {
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	(*i)->zeroChans(dT);
    }
}

//======================================  Fill data from a single frame
int
MultiDacc::fillChans(Interval Offset, Interval dT) {
    int rc = 0;
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	rc = (*i)->fillChans(Offset, dT);
    }
    return rc;
}

int
MultiDacc::getDaccIndex(const string& name) const {
    const_chan_iter i=findChannel(name); 
    if (i != mChanList.end()) {
	int index=i->getIndex();
	if(index==-1)
	    //--------------------------  Take the easy way out for now
	    return 0;
	else
	    return index;
    }
    return -1;
}

//======================================  Flush the specified data
int 
MultiDacc::flush(Interval Stride) {
    int rc = 0;
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	rc = (*i)->flush(Stride);
    }
    return rc;
}

//======================================  Open a file for input.
int
MultiDacc::seek(Time STime) {
    int rc = 0;
    for (auto i=mIn.begin(); i != mIn.end(); ++i) {
	int rci = (*i)->seek(STime);
	if (rci) rc = rci;
    }
    return rc;
}

//======================================  Get a reference to the named Channel
TSeries* 
MultiDacc::refData(const string& name) {
    chan_iter chaninx = findChannel(name);
    if (chaninx != mChanList.end()) {
	int id = chaninx->getIndex();
	if (id >= 0) {
	   if (id < mIn.size())return mIn[id]->refData(chaninx->getName());
	   throw runtime_error("MultiDacc: Invalid stream number");
	}
    }
    return (TSeries*) 0;
}

const TSeries* 
MultiDacc::refData(const string& name) const {
    auto chaninx = findChannel(name);
    if (chaninx != mChanList.end()) {
	int id = chaninx->getIndex();
	if (id >= 0) {
	   if (id < mIn.size())return mIn[id]->refData(chaninx->getName());
	   throw runtime_error("MultiDacc: Invalid stream number");
	}
    }
    return (TSeries*) 0;
}

FSeries* 
MultiDacc::refFData(const string& name) {
    chan_iter chaninx = findChannel(name);
    if (chaninx != mChanList.end()) {
	int id = chaninx->getIndex();
	if (id != -1) return mIn[id]->refFData(chaninx->getName());
    }
    return (FSeries*) 0;
}

//======================================  Synch with a time out.
int
MultiDacc::dacc_status::tseek(const Time& t0, double dt) {
    int rc = -4;
    if (!dacc_ptr) return rc;
    if (dacc_ptr->isOnline()) {
	if (!t0) {
	    if (ignore_missing_frame() && dt >= 0) dacc_ptr->setTimeout(dt);
	    rc = dacc_ptr->synch();
            dacc_ptr->setTimeout(-1.0);
        } else {
	    if (ignore_missing_frame() && dt >= 0) {
	        dacc_ptr->setDeadLine(t0 + dt);
            }
	    rc = dacc_ptr->seek(t0);
	    dacc_ptr->setDeadLine(Time(0));
	}
    } else {
	rc = dacc_ptr->seek(t0);
    }
    return rc;
}
